/*
* BSD License
*
* Copyright (c) 2007, The University of Manchester (UK)
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*     - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     - Redistributions in binary form must reproduce the above
*       copyright notice, this list of conditions and the following
*       disclaimer in the documentation and/or other materials provided
*       with the distribution.
*     - Neither the name of the University of Manchester nor the names
*       of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written
*       permission.

* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


//Simple Lee's Routing Algorithm
//Author: IW


using System;
using System.Threading;
using System.Collections;
using LeeTM.src;



public class LeeRouter
{
    virtual public WorkQueue NextTrack
    {
        get
        {
            lock (queueLock)
            {
                if (work.next != WorkQueue.nullNode)
                {
                    return work.deQueue();
                }
            }
            return null;
        }

    }
    internal const int cyan = 0x00FFFF;

    public const long MAX_SAMPLE_THRESHOLD = 60000;

    internal const int magenta = 0xFF00FF;

    internal const int yellow = 0xFFFF00;

    internal const int green = 0x00FF00;

    internal const int red = 0xFF0000;

    internal const int blue = 0x0000FF;

    //UPGRADE_NOTE: Final was removed from the declaration of 'GRID_SIZE '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal int GRID_SIZE;

    internal const int EMPTY = 0;

    internal const int TEMP_EMPTY = 10000;

    internal const int OCC = 5120;

    internal const int VIA = 6000;

    internal const int BVIA = 6001;

    internal const int TRACK = 8192;

    internal const int GOAL = 1024;

    internal const int MAX_WEIGHT = 1;

    //UPGRADE_NOTE: Final was removed from the declaration of 'grid '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal Grid grid;

    //UPGRADE_NOTE: Final was removed from the declaration of 'gridLock '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal System.Object gridLock = new System.Object();

    //private static System.Xml.XmlDocument doc;

    internal static int netNo = 0;

    //	private static int printDest = 0; // 1 for file, 0 for screen.

    // note these very useful arrays
    //UPGRADE_NOTE: Final was removed from the declaration of 'dx'. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal static readonly int[] dx = new int[] { -1, 1, 0, 0 , 0, 0, -1, 1 };

    // to help look NSEW.
    //UPGRADE_NOTE: Final was removed from the declaration of 'dy'. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal static readonly int[] dy = new int[] { 0, 0, -1, 1 , -1, 1, 0, 0 };

    // internal static Viewer view;

    internal static int failures = 0;

    internal static int num_vias = 0;

    internal static int forced_vias = 0;

    internal static System.IO.StreamReader inputFile;

    internal static System.String input_line;

    internal static int linepos = 0;

    //UPGRADE_NOTE: Final was removed from the declaration of 'queueLock '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal System.Object queueLock = new System.Object();

    //UPGRADE_NOTE: Final was removed from the declaration of 'work '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal WorkQueue work;

    //UPGRADE_NOTE: Final was removed from the declaration of 'debugQueue '. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1003'"
    internal WorkQueue debugQueue;

    public static bool TEST = false;

    public static bool DEBUG = false;

    private const bool XML_REPORT = false;


    public LeeRouter(System.String file, bool test, bool debug, bool rel)
    {
        TEST = test;
        DEBUG = debug;
        //if (DEBUG)
        //    view = new Viewer();
        if (TEST)
            GRID_SIZE = 10;
        else
            GRID_SIZE = 600;
        if (DEBUG)
            Console.WriteLine("Creating grid...");
        grid = new Grid(GRID_SIZE, GRID_SIZE, 2, rel); //the Lee 3D Grid;
        if (DEBUG)
            Console.WriteLine("Done creating grid");
        work = new WorkQueue(); // empty
        if (DEBUG)
            Console.WriteLine("Parsing data...");
        if (!TEST)
            parseDataFile(file);
        else
            fakeTestData(); //WARNING: Needs grid at least 10x10x2
        Console.printOutput();
        if (DEBUG)
            Console.WriteLine("Done parsing data");
        if (DEBUG)
            Console.WriteLine("Adding weights...");
        grid.addweights();
        if (DEBUG)
            Console.WriteLine("Done adding weights");
        Console.printOutput();
        work.sort();
        if (DEBUG)
            debugQueue = new WorkQueue();
        else
            debugQueue = null;
        Console.printOutput();
    }

    public LeeRouter(System.String file)
        : this(file, false, false, false)
    {
    }

    public LeeRouter(System.String file, bool rel)
        : this(file, false, false, rel)
    {
    }

    private void fakeTestData()
    {
        netNo++;
        grid.occupy(7, 3, 7, 3); grid.occupy(7, 7, 7, 7);
        work.next = work.enQueue(7, 3, 7, 7, netNo);

        netNo++;
        grid.occupy(3, 6, 3, 6); grid.occupy(8, 6, 8, 6);
        work.next = work.enQueue(3, 6, 8, 6, netNo);

        netNo++;
        grid.occupy(5, 3, 5, 3); grid.occupy(8, 5, 8, 5);
        work.next = work.enQueue(5, 3, 8, 5, netNo);

        netNo++;
        grid.occupy(8, 3, 8, 3); grid.occupy(2, 6, 2, 6);
        work.next = work.enQueue(8, 3, 2, 6, netNo);

        netNo++;
        grid.occupy(4, 3, 4, 3); grid.occupy(6, 7, 6, 7);
        work.next = work.enQueue(4, 3, 6, 7, netNo);

        netNo++;
        grid.occupy(3, 8, 3, 8); grid.occupy(8, 3, 8, 3);
        work.next = work.enQueue(3, 8, 8, 3, netNo);
    }

    private void parseDataFile(System.String fileName)
    {
        // Read very simple HDL file
        try
        {

            //UPGRADE_TODO: The differences in the expected value  of parameters for constructor 'java.io.BufferedReader.BufferedReader'  may cause compilation errors.  "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1092'"
            //UPGRADE_WARNING: At least one expression was used more than once in the target code. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1181'"
            //UPGRADE_TODO: Constructor 'java.io.FileInputStream.FileInputStream' was converted to 'System.IO.FileStream.FileStream' which has a different behavior. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1073_javaioFileInputStreamFileInputStream_javalangString'"
            inputFile = new System.IO.StreamReader(new System.IO.StreamReader(new System.IO.FileStream(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read), System.Text.Encoding.Default).BaseStream, new System.IO.StreamReader(new System.IO.FileStream(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read), System.Text.Encoding.Default).CurrentEncoding);
            int i = 0;
            while (true)
            {
                nextLine();
                char c = readChar();
                if (c == 'E')
                    break; // end of file
                if (c == 'C')
                // chip bounding box
                {
                    int x0 = readInt();
                    int y0 = readInt();
                    int x1 = readInt();
                    int y1 = readInt();
                    grid.occupy(x0, y0, x1, y1);
                }
                if (c == 'P')
                // pad
                {
                    int x0 = readInt();
                    int y0 = readInt();
                    grid.occupy(x0, y0, x0, y0);
                }
                if (c == 'J')
                // join connection pts
                {
                    i++;
                    int x0 = readInt();
                    int y0 = readInt();
                    int x1 = readInt();
                    int y1 = readInt();
                    netNo++;
                    work.next = work.enQueue(x0, y0, x1, y1, netNo);
                }
            }
        }
        catch (System.IO.FileNotFoundException)
        {
            Console.WriteLine("Cannot open file: " + fileName);
            System.Environment.Exit(1);
        }
        catch (System.IO.IOException exception)
        {
            //UPGRADE_TODO: Method 'java.io.PrintStream.println' was converted to 'System.Console.Out.WriteLine' which has a different behavior. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1073_javaioPrintStreamprintln_javalangObject'"
            //UPGRADE_TODO: The equivalent in .NET for method 'java.lang.Throwable.toString' may return a different value. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1043'"
            Console.WriteLine(exception);
            SupportClass.WriteStackTrace(exception, System.Console.Error);
        }
    }

    public virtual bool layNextTrack(WorkQueue q, RowMajorIntArray tempg)
    {
        // start transaction
        bool done = false;
#if STM
        try
#else
        lock (gridLock)
#endif
        {
            done = connect(q.x1, q.y1, q.x2, q.y2, q.nn, tempg, grid);
            if (DEBUG && done)
            {
                debugQueue.next = debugQueue.enQueue(q);
            }
        }
#if STM
        catch(AtomicFakeException){}
#endif
        return done;
        // end transaction
    }

    private static void nextLine()
    {
        input_line = inputFile.ReadLine();
        linepos = 0;
    }

    private static char readChar()
    {
        while ((input_line[linepos] == ' ') && (input_line[linepos] == '\t'))
            linepos++;
        char c = input_line[linepos];
        if (linepos < input_line.Length - 1)
            linepos++;
        return c;
    }

    private static int readInt()
    {
        while ((input_line[linepos] == ' ') || (input_line[linepos] == '\t'))
            linepos++;
        int fpos = linepos;
        while ((linepos < input_line.Length) && (input_line[linepos] != ' ') && (input_line[linepos] != '\t'))
            linepos++;
        int n = System.Int32.Parse(input_line.Substring(fpos, (linepos) - (fpos)));
        return n;
    }

    public virtual bool ok(int x, int y)
    {
        // checks that point is actually within the bounds
        // of grid array
        return (x > 0 && x < GRID_SIZE - 1 && y > 0 && y < GRID_SIZE - 1);
    }

    public virtual bool expandFromTo(int x, int y, int xGoal, int yGoal, int num, RowMajorIntArray tempg, Grid grid)
    {
        // this method should use Lee's expansion algorithm from
        // coordinate (x,y) to (xGoal, yGoal) for the num iterations
        // it should return true if the goal is found and false if it is not
        // reached within the number of iterations allowed.

        // g[xGoal, yGoal, 0] = EMPTY; // set goal as empty
        // g[xGoal, yGoal, 1] = EMPTY; // set goal as empty
        //UPGRADE_ISSUE: The following fragment of code could not be parsed and was not converted. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1156'"
        ArrayList_atomic front = new ArrayList_atomic();
        //UPGRADE_ISSUE: The following fragment of code could not be parsed and was not converted. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1156'"
        ArrayList_atomic tmp_front = new ArrayList_atomic();
        tempg[x, y, 0] = 1; // set grid (x,y) as 1
        tempg[x, y, 1] = 1; // set grid (x,y) as 1
        bool trace1 = false;
        front.Add(new Frontier(x, y, 0, 0));
        front.Add(new Frontier(x, y, 1, 0)); // we can start from either
        // side
        if (DEBUG)
            Console.WriteLine("Expanding " + x + " " + y + " " + xGoal + " " + yGoal);
        int extra_iterations = 50;
        bool reached0 = false;
        bool reached1 = false;
        while (front.Count != 0)
        {
            while (front.Count != 0)
            {
                int weight, prev_val;
                Frontier f = (Frontier)front.Get(0);
                front.RemoveAt(0);
                if (f.dw > 0)
                {
                    tmp_front.Add(new Frontier(f.x, f.y, f.z, f.dw - 1));
                }
                else
                {
                    if (trace1)
                        if (DEBUG)
                            Console.WriteLine("X " + f.x + " Y " + f.y + " Z " + f.z + " DW " + f.dw + " processing - val " + tempg[f.x, f.y, f.z]);
                    //					int dir_weight = 1;
                    weight = grid.getPoint(f.x, f.y + 1, f.z) + 1;
                    prev_val = tempg[f.x, f.y + 1, f.z];
                    bool reached = (f.x == xGoal) && (f.y + 1 == yGoal);
                    if ((prev_val > tempg[f.x, f.y, f.z] + weight) && (weight < OCC) || reached)
                    {
                        if (ok(f.x, f.y + 1))
                        {
                            tempg[f.x, f.y + 1, f.z] = tempg[f.x, f.y, f.z] + weight; // looking north
                            if (!reached)
                                tmp_front.Add(new Frontier(f.x, f.y + 1, f.z, 0));
                        }
                    }
                    weight = grid.getPoint(f.x + 1, f.y, f.z) + 1;
                    prev_val = tempg[f.x + 1, f.y, f.z];
                    reached = (f.x + 1 == xGoal) && (f.y == yGoal);
                    if ((prev_val > tempg[f.x, f.y, f.z] + weight) && (weight < OCC) || reached)
                    {
                        if (ok(f.x + 1, f.y))
                        {
                            tempg[f.x + 1, f.y, f.z] = tempg[f.x, f.y, f.z] + weight; // looking east
                            if (!reached)
                                tmp_front.Add(new Frontier(f.x + 1, f.y, f.z, 0));
                        }
                    }
                    weight = grid.getPoint(f.x, f.y - 1, f.z) + 1;
                    prev_val = tempg[f.x, f.y - 1, f.z];
                    reached = (f.x == xGoal) && (f.y - 1 == yGoal);
                    if ((prev_val > tempg[f.x, f.y, f.z] + weight) && (weight < OCC) || reached)
                    {
                        if (ok(f.x, f.y - 1))
                        {
                            tempg[f.x, f.y - 1, f.z] = tempg[f.x, f.y, f.z] + weight; // looking south
                            if (!reached)
                                tmp_front.Add(new Frontier(f.x, f.y - 1, f.z, 0));
                        }
                    }
                    weight = grid.getPoint(f.x - 1, f.y, f.z) + 1;
                    prev_val = tempg[f.x - 1, f.y, f.z];
                    reached = (f.x - 1 == xGoal) && (f.y == yGoal);
                    if ((prev_val > tempg[f.x, f.y, f.z] + weight) && (weight < OCC) || reached)
                    {
                        if (ok(f.x - 1, f.y))
                        {
                            tempg[f.x - 1, f.y, f.z] = tempg[f.x, f.y, f.z] + weight; // looking west
                            if (!reached)
                                tmp_front.Add(new Frontier(f.x - 1, f.y, f.z, 0));
                        }
                    }
                    if (f.z == 0)
                    {
                        weight = grid.getPoint(f.x, f.y, 1) + 1;
                        if ((tempg[f.x, f.y, 1] > tempg[f.x, f.y, 0]) && (weight < OCC))
                        {
                            tempg[f.x, f.y, 1] = tempg[f.x, f.y, 0];
                            tmp_front.Add(new Frontier(f.x, f.y, 1, 0));
                        }
                    }
                    else
                    {
                        weight = grid.getPoint(f.x, f.y, 0) + 1;
                        if ((tempg[f.x, f.y, 0] > tempg[f.x, f.y, 1]) && (weight < OCC))
                        {
                            tempg[f.x, f.y, 0] = tempg[f.x, f.y, 1];
                            tmp_front.Add(new Frontier(f.x, f.y, 0, 0));
                        }
                    }
                    // must check if found goal, if so return TRUE
                    reached0 = tempg[xGoal, yGoal, 0] != TEMP_EMPTY;
                    reached1 = tempg[xGoal, yGoal, 1] != TEMP_EMPTY;
                    if ((reached0 && !reached1) || (!reached0 && reached1))
                        extra_iterations = 100;
                    if ((extra_iterations == 0) && (reached0 || reached1) || (reached0 && reached1))
                    {
                        return true; // if (xGoal, yGoal) can be found in
                        // time
                    }
                    else
                        extra_iterations--;
                }
            };
            ArrayList_atomic tf;
            tf = front;
            front = tmp_front;
            tmp_front = tf;
        }
        //		 view.pad(x,y,red);
        //		 view.pad(xGoal,yGoal,red);
        return false;
    }

    private bool pathFromOtherSide(RowMajorIntArray g, int X, int Y, int Z)
    {
        bool ok;
        int Zo;
        Zo = 1 - Z; // other side
        int sqval = g[X, Y, Zo];
        if ((sqval == VIA) || (sqval == BVIA))
            return false;
        ok = (g[X, Y, Zo] <= g[X, Y, Z]);
        if (ok)
            ok = (g[X - 1, Y, Zo] < sqval) || (g[X + 1, Y, Zo] < sqval) || (g[X, Y - 1, Zo] < sqval) || (g[X, Y + 1, Zo] < sqval);
        return ok;
    }

    private int tlength(int x1, int y1, int x2, int y2)
    {
        int sq = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
        //UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1042'"
        return (int)System.Math.Sqrt((double)sq);
    }

    //	private static int ratio(int x1, int y1, int x2, int y2) {
    //		int xdiff = x2 - x1;
    //		int ydiff = y2 - y1;
    //		if (xdiff < 0)
    //			xdiff = -xdiff;
    //		if (ydiff < 0)
    //			ydiff = -ydiff;
    //		if (xdiff > ydiff) {
    //			if (ydiff == 0)
    //				return 100000;
    //			else
    //				return xdiff / ydiff;
    //		} else {
    //			if (xdiff == 0)
    //				return 100000;
    //			else
    //				return ydiff / xdiff;
    //		}
    //	}

    private static int deviation(int x1, int y1, int x2, int y2)
    {
        int xdiff = x2 - x1;
        int ydiff = y2 - y1;
        if (xdiff < 0)
            xdiff = -xdiff;
        if (ydiff < 0)
            ydiff = -ydiff;
        if (xdiff < ydiff)
            return xdiff;
        else
            return ydiff;
    }

    public virtual void backtrackFrom(int xGoal, int yGoal, int xStart, int yStart, int trackNo, RowMajorIntArray tempg, Grid grid)
    {
        // this method should backtrack from the goal position (xGoal, yGoal)
        // back to the starting position (xStart, yStart) filling in the
        // grid array g with the specified track number trackNo ( + TRACK).

        // ***
        // CurrentPos = Goal
        // Loop
        // Find dir to start back from current position
        // Loop
        // Keep going in current dir and Fill in track (update currentPos)
        // Until box number increases in this current dir
        // Until back at starting point
        // ***
        //		int count = 100;
        if (DEBUG)
            Console.WriteLine("Track " + trackNo + " backtrack " + "Length " + tlength(xStart, yStart, xGoal, yGoal));
        //		boolean trace = false;
        int zGoal;
        int distsofar = 0;
        if (System.Math.Abs(xGoal - xStart) > System.Math.Abs(yGoal - yStart))
            zGoal = 0;
        else
            zGoal = 1;
        if (tempg[xGoal, yGoal, zGoal] == TEMP_EMPTY)
        {
            if (DEBUG)
                Console.WriteLine("Preferred Layer not reached " + zGoal);
            zGoal = 1 - zGoal;
        }
        int tempY = yGoal;
        int tempX = xGoal;
        int tempZ = zGoal;
        int lastdir = -10;
        while ((tempX != xStart) || (tempY != yStart))
        {
            // PDL: until back

            // at starting point
            bool advanced = false;
            int mind = 0;
            int dir = 0;
            int min_square = 100000;
            int d;
            for (d = 0; d < 4; d++)
            {
                // PDL: Find dir to start back from
                // current position
                if ((tempg[tempX + dx[tempZ * 2 + d], tempY + dy[tempZ * 2 + d], tempZ] < tempg[tempX, tempY, tempZ]) && (tempg[tempX + dx[tempZ * 2 + d], tempY + dy[tempZ * 2 + d], tempZ] != TEMP_EMPTY))
                {
                    if (tempg[tempX + dx[tempZ * 2 + d], tempY + dy[tempZ * 2 + d], tempZ] < min_square)
                    {
                        min_square = tempg[tempX + dx[tempZ * 2 + d], tempY + dy[tempZ * 2 + d], tempZ];
                        mind = d;
                        dir = dx[tempZ * 2 + d] * 2 + dy[tempZ * 2 + d]; // hashed dir
                        if (lastdir < -2)
                            lastdir = dir;
                        advanced = true;
                    }
                }
            }
            if (advanced)
                distsofar++;
            if (DEBUG)
                Console.WriteLine("Backtracking " + tempX + " " + tempY + " " + tempZ + " " + tempg[tempX, tempY, tempZ] + " " + advanced + " " + mind);
            if (pathFromOtherSide(tempg, tempX, tempY, tempZ) && ((mind > 1) && (distsofar > 15) && (tlength(tempX, tempY, xStart, yStart) > 15) || (!advanced && ((grid.getPoint(tempX, tempY, tempZ) != VIA) && (grid.getPoint(tempX, tempY, tempZ) != BVIA)))))
            {
                int tZ = 1 - tempZ; // 0 if 1, 1 if 0
                int viat;
                if (advanced)
                    viat = VIA;
                else
                    viat = BVIA; // BVIA is nowhere else to go
                // mark via
                tempg[tempX, tempY, tempZ] = viat;
                grid.setPoint(tempX, tempY, tempZ, viat);
                if (DEBUG)
                    grid.setDebugPoint(tempX, tempY, tempZ, trackNo);
                tempZ = tZ;
                // and the other side
                tempg[tempX, tempY, tempZ] = viat;
                grid.setPoint(tempX, tempY, tempZ, viat);
                if (DEBUG)
                    grid.setDebugPoint(tempX, tempY, tempZ, trackNo);
                num_vias++;
                if (!advanced)
                    forced_vias++;
                if (advanced)
                    if (DEBUG)
                        Console.WriteLine("Via " + distsofar + " " + tlength(tempX, tempY, xStart, yStart) + " " + deviation(tempX, tempY, xStart, yStart));
                distsofar = 0;
            }
            else
            {
                if (grid.getPoint(tempX, tempY, tempZ) < OCC)
                {
                    // PDL: fill in track unless connection point
                    grid.setPoint(tempX, tempY, tempZ, TRACK);
                    if (DEBUG)
                        grid.setDebugPoint(tempX, tempY, tempZ, trackNo);
                }
                else if (grid.getPoint(tempX, tempY, tempZ) == OCC)
                {
                    if (DEBUG)
                        grid.setDebugPoint(tempX, tempY, tempZ, OCC);
                    if (DEBUG)
                        grid.setDebugPoint(tempX, tempY, 1 - tempZ, OCC);
                }
                tempX = tempX + dx[tempZ * 2 + mind]; // PDL: updating current
                // position on x axis
                tempY = tempY + dy[tempZ * 2 + mind]; // PDL: updating current
                // position on y axis
            }
            lastdir = dir;
        }
        if (DEBUG)
            Console.WriteLine("Track " + trackNo + " completed");
    }

    public virtual bool connect(int xs, int ys, int xg, int yg, int netNo, RowMajorIntArray tempg, Grid grid)
    {
        // calls expandFrom and backtrackFrom to create connection
        // This is the only real change needed to make the program
        // transactional.
        // Instead of using the grid 'in place' to do the expansion, we take a
        // copy
        // but the backtrack writes to the original grid.
        // This is not a correctness issue. The transactions would still
        // complete eventually without it.
        // However the expansion writes are only temporary and do not logically
        // conflict.
        // There is a question as to whether a copy is really necessary as a
        // transaction will anyway create
        // its own copy. if we were then to distinguish between writes not to be
        // committed (expansion) and
        // those to be committed (backtrack), we would not need an explicit
        // copy.
        // Taking the copy is not really a computational(time) overhead because
        // it avoids the grid 'reset' phase
        // needed if we do the expansion in place.
        // call the expansion method to return found/not found boolean
        bool found = expandFromTo(xs, ys, xg, yg, GRID_SIZE * 5, tempg, grid);
        if (found)
        {
            if (DEBUG)
                Console.WriteLine("Target (" + xg + ", " + yg + ")... FOUND!");
            backtrackFrom(xg, yg, xs, ys, netNo, tempg, grid); // call the
            // backtrack method
        }
        // print outcome of expansion method
        else
        {
            if (DEBUG)
                Console.WriteLine("Failed to route " + xs + " " + ys + " to " + xg + "  " + yg);
            failures++;
        }
       //if (DEBUG)
        //{
        //    dispGrid(grid, 0); // print the grid to screen
        //    dispGrid(grid, 1); // print the grid to screen
        //    //UPGRADE_TODO: Method 'java.awt.Component.repaint' was converted to 'System.Windows.Forms.Control.Refresh' which has a different behavior. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1073_javaawtComponentrepaint'"
        //    view.Refresh();
        //}
        return found;
    }

    public virtual void dispGrid(Grid g)
    {
    
        for (int y = 0; y < GRID_SIZE; y++){
            for (int x = 0; x < GRID_SIZE; x++)
            {
                Console.Write("|" + (g.getDebugPoint(x, y, 0) == OCC ? "x" : g.getDebugPoint(x, y, 0) + "") + "&" + (g.getDebugPoint(x, y, 1) == OCC ? "x" : g.getDebugPoint(x, y, 1) + ""));
            }
            Console.WriteLine();
        }
    }

    //public virtual void dispGrid(Grid g, int z)
    //{
    //    int laycol;
    //    if (z == 0)
    //        laycol = magenta;
    //    else
    //        laycol = green;
    //    for (int y = GRID_SIZE - 1; y >= 0; y--)
    //    {
    //        for (int x = 0; x < GRID_SIZE; x++)
    //        {
    //            int gg = g.getPoint(x, y, z);
    //            if (gg == OCC)
    //            {
    //                view.point(x, y, cyan); continue;
    //            }
    //            if (gg == VIA)
    //            {
    //                view.point(x, y, yellow); continue;
    //            }
    //            if (gg == BVIA)
    //            {
    //                view.point(x, y, red); continue;
    //            }
    //            if (gg == TRACK)
    //            {
    //                view.point(x, y, laycol); continue;
    //            } //
    //        }
    //    }
    //}


    public virtual LeeThread createThread()
    {
        return createThread(0);
    }
    public virtual LeeThread createThread(int which)
    {
        try
        {
            LeeThread leeThread = new LeeThread(this);
            return leeThread;
        }
        catch (System.Exception e)
        {
            SupportClass.WriteStackTrace(e, System.Console.Out);
            return null;
        }
    }

    public virtual void report()
    {
        //Open GUI view of PCB
        //view.display();
        //Print the PCB in ASCII, output to file
        //grid.printLayout(true);
        Console.WriteLine("Total Tracks " + netNo + " Failures " + failures + " Vias " + num_vias + " Forced Vias " + forced_vias);
    }

    public virtual void sanityCheck()
    {
        int found = 0, missing = 0;
        // Check debugGrid that the routes in debugQueue have been laid
        if (DEBUG)
        {
            System.Console.Out.WriteLine("DEBUG: Starting sanity check");
            while (debugQueue.next != WorkQueue.nullNode)
            {
                WorkQueue n = debugQueue.deQueue();
                if (!grid.findTrack(n.x1, n.y1, n.x2, n.y2, n.nn))
                {
                    System.Console.Out.WriteLine("ERROR: Missing track " + n.nn);
                    missing++;
                }
                else
                {
                    found++;
                }
            }
            System.Console.Out.WriteLine("DEBUG: found " + found + " missing " + missing);
        }
    }

    //public static void xmlReport(System.Xml.XmlDocument doc)
    //{
    //    System.Xml.XmlElement root = (System.Xml.XmlElement)doc.DocumentElement;
    //    System.Xml.XmlElement element = doc.CreateElement("BenchmarkSpecific");
    //    root.AppendChild(element);

    //    System.Xml.XmlElement total = doc.CreateElement("Tracks");
    //    total.InnerText = "0";
    //    element.AppendChild(total);

    //    System.Xml.XmlElement tracks = doc.CreateElement("Laid");
    //    tracks.InnerText = "0";
    //    element.AppendChild(tracks);

    //    System.Xml.XmlElement subnode = doc.CreateElement("Failures");
    //    subnode.InnerText = "0";
    //    element.AppendChild(subnode);
    //}

    [STAThread]
    public static void Main(System.String[] args)
    {
        if (args.Length != 2)
        {
            System.Console.Out.WriteLine("Params: [numthreads] [input-file]");
            // Console.printOutput();
            System.Environment.Exit(-1);
        }
        System.Console.Out.WriteLine("Starting");
        int numThreads = System.Int32.Parse(args[0]);
        System.String filename = args[1];
        LeeRouter lr = new LeeRouter(filename, false, false, false);
        int numMillis = 600000;

        //		 Set up the benchmark
        long startTime = 0;
        long currentTime = 0;
        long lastSample = 0;
        long maxSampleThreshold = MAX_SAMPLE_THRESHOLD;
        bool waitingForSample = false;
        long watchdogInterval = 1000;
        bool exitByTimeout = false;
        int sampleInterval = 10000;

        //if (XML_REPORT)
        //{
        //    XMLHelper.initializeXMLReport(numThreads, 100, sampleInterval, "0", "leeroutercoarse", "0");
        //}
        //else
        //{
        //Console.WriteLine("Benchmark: " + benchmarkClassName);
        //Console.WriteLine("Adapter: " + adapterClassName);
        //Console.WriteLine("Contention manager: " + managerClassName);
        System.Console.Out.WriteLine("Threads: " + numThreads);
        //Console.WriteLine("Mix: " + experiment + "% updates");
        //}

        LeeThread[] leeThreads = new LeeThread[numThreads];
        Thread[] threads = new Thread[numThreads];

        try
        {
            for (int i = 0; i < numThreads; i++)
            {
                leeThreads[i] = lr.createThread();
                threads[i] = new Thread(new ThreadStart(leeThreads[i].run));
                threads[i].Name = i + "";
            }
            startTime = (System.DateTime.Now.Ticks - 621355968000000000) / 10000;
            lastSample = startTime;
            for (int i = 0; i < numThreads; i++)
            {
                threads[i].Start();
            }
            // exitByTimeout = monitorBenchmarkToEnd(numMillis, sampleInterval, startTime, currentTime, lastSample, maxSampleThreshold, watchdogInterval, exitByTimeout, waitingForSample, leeThreads);

            // LeeThread.stop = true; // notify threads to stop
            for (int i = 0; i < numThreads; i++)
            {
                threads[i].Join();
            }
            currentTime = (System.DateTime.Now.Ticks - 621355968000000000) / 10000;
            
        }
        catch (System.Exception e)
        {
            Console.printOutput();
            SupportClass.WriteStackTrace(e, System.Console.Out);
            System.Environment.Exit(0);
            Console.printOutput();
        }

        if (lr.GRID_SIZE <= 10)
            lr.dispGrid(lr.grid);
        Console.printOutput();
        lr.sanityCheck();
        Console.printOutput();

        long elapsedTime = currentTime - startTime;
        int throughput = (int)(LeeRouter.netNo / elapsedTime);


        
        System.Console.Out.WriteLine("Numthreads: " + numThreads);
        System.Console.Out.WriteLine("Throughput:  " + throughput);
        System.Console.Out.WriteLine("ElapsedTime: " + elapsedTime);
        LeeRouter.report(startTime, exitByTimeout, true);
     
        Console.printOutput();
    }


    private static bool monitorBenchmarkToEnd(int numMillis, int sampleInterval, long startTime, long currentTime, long lastSample, long maxSampleThreshold, long watchdogInterval, bool exitByTimeout, bool waitingForSample, LeeThread[] thread)
    {
        while (!exitByTimeout)
        {
            bool exit = true;
            Thread.Sleep((int)watchdogInterval);

            // Any threads with work left?
            foreach (LeeThread i in thread)
                if (i.finished != true)
                {
                    exit = false;
                    break;
                }
            if (!exit)
                currentTime = (System.DateTime.Now.Ticks - 621355968000000000) / 10000;
            else
                break;

            // Timeout?
            if (currentTime - startTime > numMillis)
            {
                exitByTimeout = true;
            }

            // Sample?
            // Warning: watchdogInterval should remain
            // short to prevent this figures going sloppy
            if (currentTime - lastSample > sampleInterval && !waitingForSample)
            {
                waitingForSample = true;
                foreach (LeeThread t in thread)
                {
                    t.resetMyStatistics();
                    t.doneSample = false;
                    t.sampleNow = true;
                }
            }
            // Get results from all threads
            // Note: avoiding synchronisation (costs), hence
            // poor method of waiting for sample updates from threads
            // and hoping they are done by the time we go to check
            if (waitingForSample)
            {
                bool statsUpdated = true;
                foreach (LeeThread t in thread)
                {
                    if (!t.finished && !t.doneSample)
                    {
                        statsUpdated = false;
                        break;
                    }
                }

                if (statsUpdated || currentTime - lastSample > maxSampleThreshold)
                {
                    waitingForSample = false;
                    lastSample = currentTime;
                    double elapsed = (double)(lastSample - startTime) / 1000.0;
                    obtainStats(thread, elapsed, XML_REPORT);
                }
            }
        }
        return exitByTimeout;
    }

    private static void report(long startTime, bool timeout, bool xmlreport)
    {

        long stopTime = (System.DateTime.Now.Ticks - 621355968000000000) / 10000;
        //UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1042'"
        double elapsed = (double)(stopTime - startTime) / 1000.0;
        //if (xmlreport)
        //{
        //    XMLHelper.generateReportSummary(timeout, xmlreport, elapsed);
        //}
        //else
        //{

        //benchmark.report();
        Console.WriteLine("Elapsed time: " + elapsed + " seconds.");
        Console.WriteLine("----------------------------------------");
        //}
    }

    internal static void obtainStats(LeeThread[] thread, double time, bool xmlreport)
    {
        //long commits = 0;
        //long transactions = 0;
        //long commitMemRefs = 0;
        //long totalMemRefs = 0;
        long laidTracks = 0;
        long totalTracks = 0;
        System.String hardware = null;
        bool finalStats = false;

        if (thread == null)
        {
            // This means we are going to exit
            finalStats = true;
            laidTracks = LeeThread.totalLaidTracks;
        }
        else
        {
            //UPGRADE_ISSUE: The following fragment of code could not be parsed and was not converted. "ms-help://MS.VSCC.v80/dv_commoner/local/redirect.htm?index='!DefaultContextWindowIndex'&keyword='jlca1156'"
            foreach (LeeThread t in thread)
            {
                //commits += t.myCommits;
                //transactions += t.myTransactions;
                //commitMemRefs += t.myCommitMemRefs;
                //totalMemRefs += t.myTotalMemRefs;
                hardware = t.hardware;
                laidTracks += t.myLaidTracks;
                if (hardware == null)
                    hardware = "";
            }
        }

        //  if (xmlreport) {
        // XMLHelper.generateXMLIntervalSample(time, laidTracks, 0, 0, 0, hardware, finalStats); //generateXMLIntervalSampleCoarse(time,laidTracks, hardware, finalStats);
        // } else {
        Console.WriteLine("Timestamp: " + time);
        //if (transactions > 0)
        //{
        //    Console.WriteLine("Committed: " + commits);
        //    Console.WriteLine("Total transactions: " + transactions);
        //    Console.WriteLine("Percent committed: " + (100 * commits)
        //    / transactions);
        //    Console.WriteLine("Commit MemRefs: " + commitMemRefs);
        //    Console.WriteLine("Total MemRefs " + totalMemRefs);
        //    Console.WriteLine("laidTracks " + laidTracks);
        //}
        //else
        //{
        //    Console.WriteLine("No transactions executed!");
        //}
        //}
    }
}