/*
* BSD License
*
* Copyright (c) 2007, The University of Manchester (UK)
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*     - Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     - Redistributions in binary form must reproduce the above
*       copyright notice, this list of conditions and the following
*       disclaimer in the documentation and/or other materials provided
*       with the distribution.
*     - Neither the name of the University of Manchester nor the names
*       of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written
*       permission.

* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;

public class XMLHelper
{

    internal static System.Xml.XmlDocument doc;

    internal static void initializeXMLReport(int numThreads, int experiment, int sampleInterval, System.String managerClassName, System.String benchmarkClassName, System.String adapterClassName)
    {
        try
        {
            doc = new System.Xml.XmlDocument();

            System.Xml.XmlElement root = doc.CreateElement("Statistics");
            doc.AppendChild(root);

            System.Xml.XmlElement element = doc.CreateElement("Benchmark");
            element.InnerText = benchmarkClassName;
            root.AppendChild(element);

            element = doc.CreateElement("Adapter");
            element.InnerText = adapterClassName;
            root.AppendChild(element);

            element = doc.CreateElement("ContentionManager");
            element.InnerText = managerClassName;
            root.AppendChild(element);

            element = doc.CreateElement("Threads");
            element.InnerText = System.Convert.ToString(numThreads);
            root.AppendChild(element);

            element = doc.CreateElement("Mix");
            element.InnerText = System.Convert.ToString(experiment);
            root.AppendChild(element);

            element = doc.CreateElement("SampleInterval");
            element.InnerText = System.Convert.ToString(sampleInterval);
            root.AppendChild(element);

            System.String name = System.Environment.UserName;
            if (name == null)
                name = "";
            element = doc.CreateElement("Owner");
            element.InnerText = name;
            root.AppendChild(element);

            element.InnerText = System.DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");

            root.AppendChild(element);
        }
        catch (Exception e)
        {
            SupportClass.WriteStackTrace(e, Console.Error);
        }
    }

    internal static void generateReportSummary(bool timeout, bool xmlreport, double elapsed)
    {
        //try
        //{
            // get final stats
            LeeRouter.obtainStats(null, elapsed, xmlreport);

            // get benchmark to insert it's output into the doc
            // LeeRouter.xmlReport(doc);

            Console.Out.WriteLine("<ElapsedTime>" + elapsed.ToString() + "</ElapsedTime>");

            //System.Xml.XmlElement root = (System.Xml.XmlElement) doc.DocumentElement;
            //System.Xml.XmlElement element = doc.CreateElement("ElapsedTime");
            //element.InnerText = elapsed.ToString();
            //root.AppendChild(element);

            Console.Out.WriteLine("<TimeOut>" + timeout.ToString() + "</TimeOut>");
            //element = doc.CreateElement("Timeout");
            //element.InnerText = timeout.ToString();
            //root.AppendChild(element);

            // System.Xml.XmlTextWriter xmlWriter = new System.Xml.XmlTextWriter(Console.Out);

            // xmlWriter.Formatting = System.Xml.Formatting.Indented;

            // root.WriteContentTo(xmlWriter);

            // xmlWriter.Close();
            // Console.Out.WriteLine();
        //}
        //catch (System.Xml.Xsl.XsltException e)
        //{
        //    SupportClass.WriteStackTrace(e, Console.Error);
        //}
    }

    internal static void generateXMLIntervalSample(double time, long commits, long transactions, long commitMemRefs, long totalMemRefs, System.String hardware, bool finalStats)
    {
        System.Xml.XmlElement element;
        System.Xml.XmlElement root = (System.Xml.XmlElement)doc.DocumentElement;

        System.Xml.XmlElement me;
        if (finalStats)
            me = doc.CreateElement("FinalSample");
        else
            me = doc.CreateElement("Sample");
        root.AppendChild(me);
        root = me;

        element = doc.CreateElement("Timestamp");
        element.InnerText = time.ToString();
        root.AppendChild(element);

        element = doc.CreateElement("Hardware");
        element.InnerText = hardware;
        root.AppendChild(element);

        // if (transactions > 0) {
        element = doc.CreateElement("DataAvailable");
        element.InnerText = "True";
        root.AppendChild(element);

        element = doc.CreateElement("Transactions");
        element.InnerText = System.Convert.ToString(transactions);
        root.AppendChild(element);

        element = doc.CreateElement("Commits");
        element.InnerText = System.Convert.ToString(commits);
        root.AppendChild(element);

        System.String pc;
        if (commits != 0 && transactions != 0)
        {
            pc = System.Convert.ToString(100 * commits / transactions);
        }
        else
        {
            pc = "0";
        }
        element = doc.CreateElement("PercentCommits");
        element.InnerText = pc;
        root.AppendChild(element);

        element = doc.CreateElement("MemRefs");
        element.InnerText = System.Convert.ToString(totalMemRefs);
        root.AppendChild(element);

        element = doc.CreateElement("CommitMemRefs");
        element.InnerText = System.Convert.ToString(commitMemRefs);
        root.AppendChild(element);

        System.String pm;
        if (commitMemRefs != 0 && totalMemRefs != 0)
        {
            pm = System.Convert.ToString(100 * commitMemRefs / totalMemRefs);
        }
        else
        {
            pm = "0";
        }
        element = doc.CreateElement("PercentCommitMemRefs");
        element.InnerText = pm;
        root.AppendChild(element);
    }


    internal static void generateXMLIntervalSampleCoarse(double time, long tracks, System.String hardware, bool finalStats)
    {
        System.Xml.XmlElement element;
        System.Xml.XmlElement root = (System.Xml.XmlElement)doc.DocumentElement;

        System.Xml.XmlElement me;
        if (finalStats)
            me = doc.CreateElement("FinalSample");
        else
            me = doc.CreateElement("Sample");
        root.AppendChild(me);
        root = me;

        element = doc.CreateElement("Timestamp");
        element.InnerText = time.ToString();
        root.AppendChild(element);

        element = doc.CreateElement("Hardware");
        element.InnerText = hardware;
        root.AppendChild(element);

        // if (transactions > 0) {
        element = doc.CreateElement("DataAvailable");
        element.InnerText = "True";
        root.AppendChild(element);


        element = doc.CreateElement("LaidTracks");
        element.InnerText = System.Convert.ToString(tracks);
        root.AppendChild(element);
    }
}